(ns com.aurellem.gb.dv
  (:use (com.aurellem.gb gb-driver util constants))
  (:import [com.aurellem.gb.gb_driver SaveState]))

(def pokemon-1-DV-start 0xD185)

(defn pokemon-DV-start-point [poke-num]
  (+ (* poke-num pokemon-record-width)
     pokemon-1-DV-start))
   
(defn read-DV
  ([^SaveState state poke-num]
     (assert (<= 0 poke-num 5))
     (let [[raw-DV-1
            raw-DV-2]
           (subvec (vec (memory state))
                   (pokemon-DV-start-point poke-num)
                   (+ 2 (pokemon-DV-start-point poke-num)))
           defense-DV (bit-and raw-DV-1 0x0F)
           attack-DV (bit-shift-right
                       (bit-and raw-DV-1 0xF0)
                        4)
           special-DV (bit-and raw-DV-2 0x0F)
           speed-DV (bit-shift-right
                        (bit-and raw-DV-2 0xF0)
                        4)
           HP-DV
           (+
            (if (bit-test special-DV 0)   1 0)
            (if (bit-test speed-DV 0)     2 0)
            (if (bit-test defense-DV 0)   4 0)
            (if (bit-test attack-DV 0)    8 0))]
       {:attack  attack-DV
        :defense  defense-DV
        :speed  speed-DV
        :special  special-DV
        :hp  HP-DV}))
  ([poke-num]
     (read-DV @current-state poke-num)))
     

(defn give-DV
  ([^SaveState state poke-num dv-values]
     
     (assert (<= 0 poke-num 5))
     (map #(assert (<= 0 % 15)) (vals dv-values))
     (let [raw-dv-1*
           (+ (:defense dv-values)
              (bit-shift-left (:attack dv-values) 4))
           raw-dv-2*
           (+ (:special dv-values)
              (bit-shift-left (:speed dv-values) 4))
           hp-dv (:hp dv-values)
           hp-masks-1
           [[0 (bit-test hp-dv 2)]
            [4 (bit-test hp-dv 3)]]
           hp-masks-2
           [[0 (bit-test hp-dv 0)]
            [4 (bit-test hp-dv 1)]]
           set-hp-bits
           (fn [init [index hp?]]
             (if hp?
               (bit-set init index)
               (bit-clear init index)))

           raw-dv-1 (reduce set-hp-bits raw-dv-1*
                            hp-masks-1)
                     
           raw-dv-2 (reduce set-hp-bits raw-dv-2*
                            hp-masks-2)
           
           dv-start (pokemon-DV-start-point poke-num)]

       (if (or (not= raw-dv-1* raw-dv-1)
               (not= raw-dv-2* raw-dv-2))
         (println "Warning: inconsistent DV-values."
                  "Using HP settings."))
       
       (set-memory 
        (set-memory state dv-start raw-dv-1)
        (inc dv-start) raw-dv-2)))
  ([poke-num dv-values]
     (give-DV @current-state poke-num dv-values)))

(def good-DVs
  {:attack  15
   :defense 15
   :speed   15
   :special 15
   :hp      15})

(def bad-DVs
  {:attack  0
   :defense 0
   :speed   0
   :special 0
   :hp      0})
